# Module:   __init__
# Date:     3rd October 2008
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""sahriswiki - Lightweight Wiki Engine

:copyright: CopyRight (C) 2010-2011 by James Mills
:license: MIT (See: LICENSE)
"""

__author__ = "James Mills"
__date__ = "31st January 2011"
__version__ = "0.8"



# Module:   unrepr
# Date:     15th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au
# NOTE:     Borrowed from CherryPy3

"""unrepr

unrepr(...) function and support borrowed from the CherryPy3 library
with fixes and improvements.
"""

# public domain "unrepr" implementation, found on the web and then improved.

import sys
import operator as _operator

class _Builder:
    
    def build(self, o):
        m = getattr(self, 'build_' + o.__class__.__name__, None)
        if m is None:
            raise TypeError("unrepr does not recognize %s" %
                            repr(o.__class__.__name__))
        return m(o)
    
    def build_Subscript(self, o):
        return self.build(o.value)[self.build(o.slice)]
    
    def build_Index(self, o):
        return self.build(o.value)
    
    def build_Call(self, o):
        callee = self.build(o.func)
        
        if o.args is None:
            args = ()
        else: 
            args = tuple([self.build(a) for a in o.args]) 
        
        if o.starargs is None:
            starargs = ()
        else:
            starargs = self.build(o.starargs)
        
        if o.kwargs is None:
            kwargs = {}
        else:
            kwargs = self.build(o.kwargs)
        
        return callee(*(args + starargs), **kwargs)
    
    def build_List(self, o):
        return list(map(self.build, o.elts))
    
    def build_Str(self, o):
        return o.s
    
    def build_Num(self, o):
        return o.n
    
    def build_Dict(self, o):
        return dict([(self.build(k), self.build(v))
                     for k, v in zip(o.keys, o.values)])
    
    def build_Tuple(self, o):
        return tuple(self.build_List(o))
    
    def build_Name(self, o):
        name = o.id
        if name == 'None':
            return None
        if name == 'True':
            return True
        if name == 'False':
            return False
        
        # See if the Name is a package or module. If it is, import it.
        try:
            return modules(name)
        except ImportError:
            pass
        
        # See if the Name is in __builtins__.
        try:
            return getattr(__builtins__, name)
        except AttributeError:
            pass
        
        raise TypeError("unrepr could not resolve the name %s" % repr(name))
    
    def build_BinOp(self, o):
        left, op, right = map(self.build, [o.left, o.op, o.right]) 
        return op(left, right)

    def build_Add(self, o):
        return _operator.add

    def build_Attribute(self, o):
        parent = self.build(o.value)
        return getattr(parent, o.attr)

    def build_NoneType(self, o):
        return None


def _astnode(s):
    """Return a Python ast Node compiled from a string."""
    try:
        import ast
    except ImportError:
        # Fallback to eval when ast package is not available,
        # e.g. IronPython 1.0.
        return eval(s)

    p = ast.parse("__tempvalue__ = " + s)
    return p.body[0].value

def unrepr(s):
    """Return a Python object compiled from a string."""
    if not s:
        return s
    obj = _astnode(s)
    return _Builder().build(obj)


def modules(modulePath):
    """Load a module and retrieve a reference to that module."""
    try:
        mod = sys.modules[modulePath]
        if mod is None:
            raise KeyError()
    except KeyError:
        # The last [''] is important.
        mod = __import__(modulePath, globals(), locals(), [''])
    return mod



# Module:   html
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""HTML Macros

Macros for generating snippets of HTML.
"""

from genshi.input import HTML
from genshi.core import Markup
from genshi.builder import tag
from genshi.filters import HTMLSanitizer
from genshi.output import HTMLSerializer

sanitizer = HTMLSanitizer()
serializer = HTMLSerializer()

from sahriswiki.highlight import highlight

def code(macro, environ, data, *args, **kwargs):
    """Displays a block of text with syntax highlighting in a HTML <pre>.
    
    This macro uses the pygments syntax highlighter to render a block of
    text given a specific language. For a list of available languages
    that are supported by pygments see:
    * [[http://pygments.org/languages/]]

    If an invalid language is specified or a lexer cannot be found for the
    given language (//given by the lang argument//) then the block of text
    will be rendered unhighlighted.

    **Arguments:**
    * lang=None (//the language to highlight//)

    **Example(s):**
    {{{
    <<code lang="python">>
    def main():
        print "Hello World!"

    if __name__ == "__main__":
        main()
    <</code>>
    }}}

    <<code lang="python">>
    def main():
        print "Hello World!"

    if __name__ == "__main__":
        main()
    <</code>>
    """
    
    if not macro.body:
        return None

    lang = kwargs.get("lang", None)
    linenos = kwargs.get("linenos", False)

    return highlight(macro.body, lang=lang, linenos=linenos)

def div(macro, environ, data, *args, **kwargs):
    """Displays a block of text in a custom HTML <div>.
    
    This macro allows you to render a block of text in a custom HTML <div>
    block and contrib various attributes such as style, class, id, etc.

    **Arguments:**
    * id=None (//the id attribute//)
    * class=None (//the class attribute//)
    * float=None (//the float style//)
    * style=None (//the style attribute//)
    * context="block" (//Either "inline" or "block"//)

    **Example(s):**
    {{{
    <<div float="right">>
    Hello World!
    <</div>>
    }}}

    <<div float="right">>
    Hello World!
    <</div>>
    """

    if macro.body is None:
        return None

    id = kwargs.get("id", "")
    cls = kwargs.get("class", "")
    float = kwargs.get("float", "")
    style = kwargs.get("style", "")
    parse = kwargs.get("parse", True)
    context = kwargs.get("context", "block")

    if float and float in ("left", "right"):
        style = "float: %s; %s" % (float, style)

    if style:
        style = ";".join(sanitizer.sanitize_css(style))

    contents = environ.parser.generate(macro.body, context=context,
            environ=(environ, data))

    attrs = {}
    if id:
        attrs["id"] = id
    if cls:
        attrs["class_"] = cls
    if style:
        attrs["style"] = style

    return tag.div(contents, **attrs)

def span(macro, environ, data, *args, **kwargs):
    """Displays text in a HTML <span>

    This macro allows you to display text in a custom HTML <span>
    element and contrib various attributes such as style, class, id, etc.
    
    **Arguments:**
    * id=None (//the id attribute//)
    * class=None (//the class attribute//)
    * style=None (//the style attribute//)

    **Example(s):**
    {{{
    <<span "Hello World!">>
    }}}

    <<span "Hello World!">>
    """

    text = macro.body or (args and args[0]) or None

    if not text:
        return None

    id = kwargs.get("id", None)
    cls = str(kwargs.get("class", None))
    style = kwargs.get("style", None)

    if style:
        style = ";".join(sanitizer.sanitize_css(style))

    contents = environ.parser.generate(text, context="inline",
            environ=(environ, data))

    attrs = {}
    if id:
        attrs["id"] = id
    if cls:
        attrs["class_"] = cls
    if style:
        attrs["style"] = style

    return tag.span(contents, **attrs)

def p(macro, environ, data, *args, **kwargs):
    """Displays text in a HTML <p>

    This macro allows you to display text in a custom HTML <p>
    element and contrib various attributes such as style, class, id, etc.
    
    **Arguments:**
    * id=None (//the id attribute//)
    * class=None (//the class attribute//)
    * style=None (//the style attribute//)

    **Example(s):**
    {{{
    <<p "Hello World!">>
    }}}

    <<p "Hello World!">>
    """

    text = macro.body or (args and args[0]) or None

    if not text:
        return None

    id = kwargs.get("id", None)
    cls = str(kwargs.get("class", None))
    style = kwargs.get("style", None)

    if style:
        style = ";".join(sanitizer.sanitize_css(style))

    contents = environ.parser.generate(text, context="inline",
            environ=(environ, data))

    attrs = {}
    if id:
        attrs["id"] = id
    if cls:
        attrs["class_"] = cls
    if style:
        attrs["style"] = style

    return tag.p(contents, **attrs)

def html(macro, environ, data, *args, **kwargs):
    """Displays raw HTML content.

    This macro allows you to display raw HTML with any **safe** content.

    **NB:** Any elements considered unsafe are automatically stripped.
    
    **Arguments:** //No Arguments//

    **Example(s):**
    {{{
    <<html>>
    <h1>Hello World!</h1>
    <</html>>
    }}}

    <<html>>
    <h1>Hello World!</h1>
    <</html>>
    """

    if not macro.body:
        return None

    return Markup("".join(serializer(sanitizer(HTML(macro.body)))))

def img(macro, environ, data, *args, **kwargs):
    """Displays an image, HTML <img>.

    This macro allows you to display a custom image, HTML <img>
    element and contrib various attributes such as style, class, id, etc.
    
    **Arguments:**
    * id=None (//the id attribute//)
    * alt=None (//the alt attribute//)
    * class=None (//the class attribute//)
    * style=None (//the style attribute//)

    **Example(s):**
    {{{
    <<img "/img/python.png">>
    }}}

    <<img "/img/python.png">>
    """

    if not args:
        return None

    src = args[0]

    id = kwargs.get("id", None)
    alt = kwargs.get("alt", None)
    cls = kwargs.get("class", None)
    style = kwargs.get("style", None)

    if style:
        style = ";".join(sanitizer.sanitize_css(style))

    attrs = {}
    if id:
        attrs["id"] = id
    if alt:
        attrs["alt"] = alt
    if cls:
        attrs["class_"] = cls
    if style:
        attrs["style"] = style

    return tag.img(src=src, **attrs)



# Module:   auth
# Date:     25th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Authentication / Permissions

...
"""

import schema

class Permissions(object):

    def __init__(self, environ, username):
        super(Permissions, self).__init__()

        self.environ = environ
        self.username = username

        self._actions = []

        db = self.environ.dbm.session
        self._actions = [permission.action \
                for permission in db.query(schema.Permission).\
                filter(schema.Permission.username==self.username)]

    def __contains__(self, action):
        return action in self._actions or "SAHRIS_ADMIN" in self._actions

    def __repr__(self):
        return "<Permissions(%s %r)>" % (self.username, self._actions)



# Module:   wiki
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Wiki Macros

Various wiki and page macros.
"""

import re
import time

from genshi.builder import tag
from genshi.core import Markup
from genshi.output import HTMLSerializer

serializer = HTMLSerializer()

def SetTitle(macro, environ, data, *args, **kwargs):
    """Set the title of the page.
    
    This macro allows you to set a custom title for a page that is
    different to the name of the page (//for SEO purposes//). You can
    also optinoally specify that you also want the newly set title
    to be rendered in-place.

    **Arguments:**
    * title (//the new title//)
    * display=False (//whether to render the title//)

    **Example(s):**
    {{{
    <<SetTitle "My Title">>
    }}}

    <<SetTitle "SahrisWiki Macros">>

    {{{
    <<SetTitle "My Title", display=True>>
    }}}

    <<SetTitle "SahrisWiki Macros", display=True>>
    """

    title = kwargs.get("title", (args and args[0]) or u"")
    display = kwargs.get("display", False)

    if not title:
        return None

    data["title"] = title

    if display:
        return title
    else:
        return Markup("<!-- %s -->" % title)

def title(macro, environ, data, *args, **kwargs):
    """Displays the current title of the page or it's name.
    
    This macro simply displays the title or name of the current page.

    **Arguments:** //No Arguments//

    **Example(s):**
    {{{
    <<title>>
    }}}

    <<title>>
    """

    if "title" in data:
        title = data["title"]
    elif "page" in data:
        title = data["page"].get("name", u"")
    else:
        title = u""

    return title

def AddComment(macro, environ, data, *args, **kwargs):
    """Display an add comment form allowing users to post comments.

    This macro allows you to display an add comment form on the current
    page allowing users to post comments. The comments are added to the
    page's content itself.
    
    **Arguments:** //No Arguments//

    **Example(s):**
    {{{
    <<AddComment>>
    }}}

    <<AddComment>>
    """

    # Setup info and defaults
    parser = environ.parser
    request = environ.request

    page = data["page"]
    page_name = page["name"]
    page_text = page["text"]
    
    # Get the data from the POST
    comment = request.kwargs.get("comment", "")
    action = request.kwargs.get("action", "")
    author = request.kwargs.get("author", environ._user())
    
    # Ensure <<add-comment>> is not present in comment, so that infinite
    # recursion does not occur.
    comment = re.sub("(^|[^!])(\<\<add-comment)", "\\1!\\2", comment)
    
    the_preview = None
    the_comment = None

    # If we are submitting or previewing, inject comment as it should look
    if action == "preview":
        the_preview = tag.div(tag.h1("Preview"), id="preview")
        the_preview += tag.div(parser.generate(comment, context="inline",
            environ=(environ, data)), class_="article")

    # When submitting, inject comment before macro
    if comment and action == "save":
        new_text = ""
        comment_text = "==== Comment by %s on %s ====\n%s\n\n" % (
                author, time.strftime('%c', time.localtime()), comment)
        for line in page_text.split("\n"):
            if line.find("<<add-comment") == 0:
                new_text += comment_text
            new_text += line + "\n"

        search = environ.search
        storage = environ.storage

        storage.reopen()
        search.update(environ)

        storage.save_text(page_name, new_text, author,
                "Comment added by %s" % author)

        search.update_page(environ.get_page(page_name), page_name,
                text=new_text)

        the_comment = parser.generate(comment_text, context="inline",
                environ=(environ, data))

    the_form = tag.form(
            tag.input(type="hidden", name="parent", value=page["node"]),
            tag.fieldset(
                tag.legend("Add Comment"),
                tag.p(
                    tag.textarea(
                        (not action in ("cancel", "save") and comment or ""),
                        id="comment",
                        name="comment",
                        cols=80, rows=5
                    ),
                    class_="text"
                ),
                tag.h4(tag.label("Your email or username:", for_="author")),
                tag.p(
                    tag.input(id="author", name="author", type="text"),
                    class_="input"
                ),
                tag.p(
                    tag.button("Preview", type="submit",
                        name="action", value="preview"),
                    tag.button("Save", type="submit",
                        name="action", value="save"),
                    tag.button("Cancel", type="submit",
                        name="action", value="cancel"),
                    class_="button"
                ),
            ),
            method="post", action=""
    )

    return tag(the_preview, the_comment, the_form)

def source(macro, environ, data, *args, **kwargs):
    """Display the HTML source of some parsed wiki text
    
    This macro allows you to display the genereated HTML source of some
    parsed wiki text. (//Useful mostly for debugging//).

    **Arguments:** //No Arguments//

    **Example(s):**
    {{{
    <<source>>**Hello World!**<</source>>
    }}}

    <<source>>**Hello World!**<</source>>
    """

    if not macro.body:
        return None

    contents = environ.parser.generate(macro.body, context="inline",
            environ=(environ, data))
    
    return tag.pre("".join(serializer(contents)))



# __init__.py
# -*- coding: utf-8 -*-
#
# Copyright © Stephen Day
#
# This module is part of Creoleparser and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
#
import string
import keyword

from core import Parser, ArgParser
from dialects import (creole11_base, creole10_base, creepy10_base,
                    create_dialect)

__docformat__ = 'restructuredtext en'

__version__ = '0.7.2'

creole2html = Parser(dialect=create_dialect(creole10_base), method='html')
"""This is a pure Creole 1.0 parser created for convenience"""

text2html = Parser(dialect=create_dialect(creole11_base), method='html')
"""This is a Creole 1.0 parser (+ additions) created for convenience"""

parse_args = ArgParser(dialect=creepy10_base(),key_func=string.lower,
                       illegal_keys=keyword.kwlist + ['macro_name',
                         'arg_string', 'body', 'isblock', 'environ', 'macro'])
"""Function for parsing macro arg_strings using a relaxed xml style"""

def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



# dialects.py
# -*- coding: utf-8 -*-
#
# Copyright © Stephen Day
#
# This module is part of Creoleparser and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
#

import warnings
import string

from elements import *



def create_dialect(dialect_base, **kw_args):
    """Factory function for dialect objects (for parameter defaults,
    see :func:`~creoleparser.dialects.creole10_base`)

    :parameters:
      dialect_base
        The class factory to use for creating the dialect object.
        ``creoleparser.dialects.creole10_base`` and  
        ``creoleparser.dialects.creole11_base`` are possible values.
      wiki_links_base_url
        The page name found in wiki links will be smartly appended to this to
        form the href. To use a different base url for images, supply a two
        element list; the second element will be used.
      wiki_links_space_char
        When wiki_links have spaces, this character replaces those spaces in
        the url. To use a different character for images, supply a two element
        list; the second element will be used.
      wiki_links_class_func
        If supplied, this fuction will be called when a wiki link is found and
        the return value (should be a string) will be added as a class attribute
        of the corresponding link. The function must accept the page name (any
        spaces will have been replaced) as it's only argument.
        If no class attribute is to be added, return `None`.
      wiki_links_path_func
        If supplied, this fuction will be called when a wiki link is found and
        the return value (should be a string) will be joined to the base_url
        to form the url for href. The function must accept the page name (any
        spaces will have been replaced) as it's only argument. Special characters
        should be url encoded. To use a different function for images, supply a
        two element list; the second element will be used.
      interwiki_links_base_urls
        Dictionary of urls for interwiki links.
      interwiki_links_space_chars
        Dictionary of characters that that will be used to replace spaces
        that occur in interwiki_links. If no key is present for an interwiki
        name, the wiki_links_space_char will be used.
      interwiki_links_funcs
        Dictionary of functions that will be called for interwiki link
        names. Works like wiki_links_path_func
      no_wiki_monospace
        If `False`, inline no_wiki will be rendered as <span> not <code>
      macro_func
        If supplied, this fuction will be called when macro markup is found. The
        function must accept the following postional arguments:
        
        1. macro name (string)
        2. the argument, including any delimter (string)
        3. the macro body (string or None for a macro without a body)
        4. macro type (boolean, True for block macros, False for normal macros)
        5. an `environ` object (see :meth:`creoleparser.core.Parser.generate`)
        
        The function may return a string (which will be subject to further wiki
        processing) or a Genshi object (Stream, Markup, builder.Fragment, or
        builder.Element). If None is returned, the markup will
        be rendered unchanged.
      blog_style_endings
        If `True`, each newline character in a paragraph will be converted to
        a <br />. Note that the escaping mechanism (tilde) does not work
        for newlines.
    """

    return dialect_base(**kw_args)



def creole10_base(wiki_links_base_url='',wiki_links_space_char='_',
                 interwiki_links_base_urls={},
                 no_wiki_monospace=True,
                 wiki_links_class_func=None,
                 wiki_links_path_func=None, interwiki_links_funcs={},
                 interwiki_links_space_chars={},
                 blog_style_endings=False,
                 ):
    """Returns a base class for extending
    (for parameter descriptions, see :func:`~creoleparser.dialects.create_dialect`)

    The returned class does not implement any of the proposed additions to
    to Creole 1.0 specification.

    """
    
    if isinstance(wiki_links_base_url,(list, tuple)):
        wiki_links_base_url, embed_base_url = wiki_links_base_url
    else:
        embed_base_url = wiki_links_base_url
    if isinstance(wiki_links_path_func,(list, tuple)):
        wiki_links_path_func, embed_path_func = wiki_links_path_func
    else:
        embed_path_func = wiki_links_path_func
    if isinstance(wiki_links_space_char,(list, tuple)):
        wiki_links_space_char, embed_space_char = wiki_links_space_char
    else:
        embed_space_char = wiki_links_space_char
        

    url_end_pattern = r'\S+?(?=([>)}\]]?[,.?!:;"\']?(([^a-zA-Z0-9])\5)?(\s|$))|<<<)'

    def raw_link_tag(mo, environ):
        if mo.group('protocol') in ['http','https']:
            tag = 'a'
        else:
            tag = ''
        return tag
    
    class Base(Dialect):

        #br = LineBreak('br', r'\\',blog_style=blog_style_endings)
        headings = Heading(['h1','h2','h3','h4','h5','h6'],'=')
        no_wiki = NoWikiElement(no_wiki_monospace and 'code' or 'span',['{{{','}}}'])
        #img = Image('img',('{{','}}'),delimiter='|')
        simple_element = SimpleElement(token_dict={'**':'strong','//':'em'})
        hr = LoneElement('hr','----')
        blank_line = BlankLine()
        p = Paragraph('p')
        pre = PreBlock('pre',['{{{','}}}'])
        #raw_link = RawLink('a',)
        br = GenericElement(blog_style_endings and r'\\\\\n?|\n(?!$)' or r'\\\\','br')
        raw_link = GenericElement('(?P<protocol>https?|ftp)://'+ url_end_pattern, raw_link_tag,'{all}', dict(href='{all}'))
        #masked_raw_link = GenericElement('(ftp)://'+ url_end_pattern,'','{all}')

        link = AnchorElement('a',('[[',']]'),delimiter = '|',interwiki_delimiter=':',
                                            base_urls=interwiki_links_base_urls,
                                            links_funcs=interwiki_links_funcs,
                                            default_space_char=wiki_links_space_char,
                                            space_chars=interwiki_links_space_chars,
                                       base_url=wiki_links_base_url,
                              space_char=wiki_links_space_char,class_func=wiki_links_class_func,
                              path_func=wiki_links_path_func)

        img = ImageElement('img',('{{','}}'),delimiter = '|',interwiki_delimiter=':',
                                            base_urls=interwiki_links_base_urls,
                                            links_funcs=interwiki_links_funcs,
                                            default_space_char=embed_space_char,
                                            space_chars=interwiki_links_space_chars,
                                       base_url=embed_base_url,
                              space_char=embed_space_char,class_func=wiki_links_class_func,
                              path_func=embed_path_func)        


        td = TableCell('td','|')
        th = TableCell('th','|=')
        tr = TableRow('tr','|')
        table = Table('table','|')

        li = ListItem('li',list_tokens='*#')
        ol = List('ol','#',stop_tokens='*')
        ul = List('ul','*',stop_tokens='#')
        nested_ol = NestedList('ol','#')
        nested_ul = NestedList('ul','*')

        def __init__(self):
            self.link.child_elements = [self.simple_element]
            self.simple_element.child_elements = [self.simple_element]
            self.headings.child_elements = self.inline_elements
            self.p.child_elements = self.inline_elements
            self.td.child_elements = self.inline_elements[3:]#[self.br, self.raw_link, self.masked_raw_link, self.simple_element]
            self.th.child_elements = self.inline_elements[3:]#[self.br, self.raw_link, self.masked_raw_link, self.simple_element]
            self.tr.child_elements = [self.no_wiki,self.img,self.link,self.th,self.td]
            self.table.child_elements = [self.tr]
            self.ol.child_elements = [self.li]
            self.ul.child_elements = [self.li]
            self.nested_ol.child_elements = [self.li]
            self.nested_ul.child_elements = [self.li]
            self.li.child_elements = [(self.nested_ol,self.nested_ul)] + self.inline_elements

        @property 
        def inline_elements(self):
            return [self.no_wiki, self.img, self.link, 
                    self.br, self.raw_link, self.simple_element]

        @property 
        def block_elements(self):
            return [self.pre,self.blank_line,self.table,self.headings,
                               self.hr,self.ul,self.ol,self.p]
            """self.block_elements are the wiki elements that are searched at the top level of text to be
            processed. The order matters because elements later in the list need not have any
            knowledge of those before (as those were parsed out already). This makes the
            regular expression patterns for later elements very simple.
            """        

    return Base



def creole11_base(macro_func=None,
                  indent_class=None,
                  indent_style='margin-left:2em',
                  **kwargs):
    r"""Returns a base class for extending (for parameter descriptions, see :func:`~creoleparser.dialects.create_dialect`)

    The returned class implements most of the *officially* proposed additions to
    to Creole 1.0 specification:

        * superscript, subscript, underline, and monospace
        * definition lists
        * indentation
        * macros
            
        (see http://purl.oclc.org/creoleparser/cheatsheet)

   **A Basic Extending Example**

   Here we create a dialect that alters the basic Creole inline syntax by
   removing underline and adding strike-though::

       >>> Base = creole11_base()
       >>> class MyDialect(Base):
       ...      simple_element = SimpleElement(token_dict={'**':'strong',
       ...                                                  '//':'em',
       ...                                                  ',,':'sub',
       ...                                                  '^^':'sup',
       ...                                                  '--':'del',
       ...                                                  '##':'code'})
       ...
       ...      my_custom_element = GenericElement('issue (?P<id>\d+)',
       ...             'a','{all}',dict(href='http://code.google.com/p/creoleparser/issues/detail?id={id}',
       ...              class_='external-link'))
       ...
       ...      @property
       ...      def inline_elements(self):
       ...          elements = super(MyDialect,self).inline_elements
       ...          elements.append(self.my_custom_element)
       ...          return elements
       ...
       >>> from core import Parser
       >>> parser = Parser(MyDialect) 
       >>> print parser.render("delete --this-- but don't underline __this__"),
       <p>delete <del>this</del> but don't underline __this__</p>
       >>> print parser.render("see issue 43 for more information"),
       <p>see <a class="external-link" href="http://code.google.com/p/creoleparser/issues/detail?id=43">issue 43</a> for more information</p>


  While extending the base class in a conventional way is sometimes needed (e.g., when extending a inline_elements()),
  in many cases, like this one, class attributes can be altered directly instead:
  
       >>> MyDialect = creole11_base()
       >>> MyDialect.simple_element = SimpleElement(token_dict={'**':'strong',
       ...                                                  '//':'em',
       ...                                                  ',,':'sub',
       ...                                                  '^^':'sup',
       ...                                                  '--':'del',
       ...                                                  '##':'code'})
       >>> parser = Parser(MyDialect) 
       >>> print parser.render("delete --this-- but don't underline __this__"),
       <p>delete <del>this</del> but don't underline __this__</p>


           
   For a more complex example, see the `source code
   <http://code.google.com/p/creoleparser/source/browse/trunk/creoleparser/dialects.py>`_
   of this function. It extends the class created from creole10_base().

   .. note::

       It is generally safest to create only one dialect instance per base
       class. This is because WikiElement objects are bound as class
       attributes and would therefor be shared between multiple instances,
       which could lead to unexpected behaviour.

    
    """
    
    Creole10Base = creole10_base(**kwargs)
    
    class Base(Creole10Base):
        
        simple_element = SimpleElement(token_dict={'**':'strong','//':'em',',,':'sub',
                                                  '^^':'sup','__':'u','##':'code'})
        indented = IndentedBlock('blockquote','>')
        
        dd = DefinitionDef('dd',':')
        dt = DefinitionTerm('dt',';',stop_token=':')
        dl = List('dl',';',stop_tokens='*#')

        macro = Macro('',('<<','>>'),func=macro_func)
        bodiedmacro = BodiedMacro('',('<<','>>'),func=macro_func)
        bodied_block_macro = BodiedBlockMacro('',('<<','>>'),func=macro_func)    

        def __init__(self):
            super(Base,self).__init__()
            self.tr.child_elements[0] = (self.no_wiki,self.bodiedmacro,self.macro)
            self.dd.child_elements = self.inline_elements[3:]
            self.dt.child_elements = self.inline_elements[3:]
            self.dl.child_elements = [(self.no_wiki,self.bodiedmacro,self.macro),self.img,self.link,self.dt,self.dd]
            self.indented.child_elements = self.block_elements
            
        @property 
        def inline_elements(self):
            elements = super(Base,self).inline_elements
            elements[0] = (self.no_wiki,self.bodiedmacro,self.macro)
            return elements

        @property 
        def block_elements(self):
            elements = super(Base,self).block_elements
            elements[0] = (self.bodied_block_macro,self.pre)
            return elements[:5] + [self.indented, self.dl] + elements[5:]

    return Base


class Dialect(object):
    """Base class for dialect objects."""
    pass



def creepy10_base():
    """Returns a dialect object (a class) to be used by :class:`~creoleparser.core.ArgParser`


    **How it Works**

    The "Creepy" dialect uses a syntax that can look much like that of
    attribute definition in xml. The most important differences are that
    positional arguments are allowed and quoting is optional.

    A Creepy dialect object is normally passed to
    :class:`~creoleparser.core.ArgParser` to create a new parser object.
    When called with a single argument, this outputs a two-tuple
    (a list of positional arguments and a dictionary of keyword arguments):

    >>> from core import ArgParser
    >>> my_parser = ArgParser(dialect=creepy10_base(), convert_implicit_lists=False)
    >>> my_parser(" foo='one' ")
    ([], {'foo': 'one'})
    >>> my_parser("  'one' ")
    (['one'], {})
    >>> my_parser("  'one' foo='two' ")
    (['one'], {'foo': 'two'})

    Positional arguments must come before keyword arguments. If they occur
    after a keyword argument, they will be combined with that value as a list:
    
    >>> my_parser("  foo='one' 'two' ")
    ([], {'foo': ['one', 'two']})

    Similarly, if two or more keywords are the same, the values will be combined
    into a list:

    >>> my_parser("  foo='one' foo='two' ")
    ([], {'foo': ['one', 'two']})

    The lists above are known as "Implicit" lists. They can automatically be
    converted to strings by setting ``convert_implicit_lists=True`` in the
    parser.

    Quotes can be single or double:
    
    >>> my_parser(''' foo="it's okay" ''')
    ([], {'foo': "it's okay"})

    Tildes can be used for escaping:

    >>> my_parser(''' foo='it~'s okay' ''')
    ([], {'foo': "it's okay"})
    
    Quotes are optional if an argument value doesn't contain spaces or
    unescaped special characters:
    
    >>> my_parser("  one foo = two ")
    (['one'], {'foo': 'two'})

    Keyword arguments lacking a value will be interpreted as an empty string:

    >>> my_parser(" '' foo=  boo= '' ")
    ([''], {'foo': '', 'boo': ''})

    """
    
    class Base(ArgDialect):

       kw_arg = KeywordArg(token='=')
       quoted_arg = QuotedArg(token='\'"')
       spaces = WhiteSpace()

       def __init__(self):
          self.kw_arg.child_elements = [self.spaces]
          self.quoted_arg.child_elements = []
          self.spaces.child_elements = []

       @property
       def top_elements(self):
          return [self.quoted_arg, self.kw_arg, self.spaces]

    return Base


def creepy20_base():
    """Extends creepy10_base to support an explicit list argument syntax.

    >>> from core import ArgParser
    >>> my_parser = ArgParser(dialect=creepy20_base(),convert_implicit_lists=False)
    >>> my_parser(" one [two three] foo=['four' 'five'] ")
    (['one', ['two', 'three']], {'foo': ['four', 'five']})

    You can test if a list is explicit by testing its class:

    >>> from core import ImplicitList
    >>> pos, kw = my_parser("  foo=['one' 'two'] boo = 'three' 'four'")
    >>> print kw
    {'foo': ['one', 'two'], 'boo': ['three', 'four']}
    >>> isinstance(kw['foo'], ImplicitList)
    False
    >>> isinstance(kw['boo'], ImplicitList)
    True
    
    Lists of length zero or one are **never** of type ImplicitList.

    """

    Creepy10Base = creepy10_base()
    class Base(Creepy10Base):

       list_arg = ListArg(token=['[',']'])
       explicit_list_arg = ExplicitListArg(token=['[',']'])

       def __init__(self):
          super(Base,self).__init__()
          self.kw_arg.child_elements = [self.explicit_list_arg,self.spaces]
          self.list_arg.child_elements = [self.spaces]
          self.explicit_list_arg.child_elements = [self.spaces]

       @property
       def top_elements(self):
          return [self.quoted_arg, self.kw_arg, self.list_arg,self.spaces]

    return Base



class ArgDialect(object):
    """Base class for argument string dialect objects."""
    pass


 
def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()    



# core.py
# -*- coding: utf-8 -*-
#
# Copyright © Stephen Day
#
# This module is part of Creoleparser and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
#

import re, string
import warnings

import genshi.builder as bldr

__docformat__ = 'restructuredtext en'

escape_char = '~'
esc_neg_look = '(?<!' + re.escape(escape_char) + ')'
esc_to_remove = re.compile(''.join([r'(?<!',re.escape(escape_char),')',re.escape(escape_char),r'(?!([ \n]|$))']))
place_holder_re = re.compile(r'<<<(-?\d+?)>>>')


class Parser(object):
    
    def __init__(self,dialect, method='xhtml', strip_whitespace=False, encoding='utf-8'):
        """Constructor for Parser objects

        :parameters:
          dialect
            Usually created using :func:`creoleparser.dialects.create_dialect`
          method
            This value is passed to Genshies Steam.render(). Possible values
            include ``xhtml``, ``html``, ``xml``, and ``text``.
          strip_whitespace
            This value is passed to Genshies Steam.render().
          encoding
            This value is passed to Genshies Steam.render(). If ``None``, the ouput
            will be a unicode object.
            
        """
    
        if isinstance(dialect,type):
            self.dialect = dialect()
        else:
            warnings.warn("""
'dialect' should be a type object.
"""
                  )
            self.dialect = dialect
        self.method = method
        self.strip_whitespace = strip_whitespace
        self.encoding=encoding


    def parse(self,text,element_store=None,context='block', environ=None, preprocess=True):
        """Returns a Genshi Fragment (basically a list of Elements and text nodes).

        :parameters:
          text
            The text to be parsed.
          context
            This is useful for marco development where (for example) supression
            of paragraph tags is desired. Can be 'inline', 'block', or a list
            of WikiElement objects (use with caution).
          element_store
            Internal dictionary that's passed around a lot ;)
          environ
            This can be any type of object. It will be passed to ``macro_func``
            unchanged (for a description of ``macro_func``, see
            :func:`~creoleparser.dialects.create_dialect`).
          preprocess
            Passes text through preprocess method that replaces Windows style
            line breaks.
            
        """
        
        if element_store is None:
            element_store = {}
        if not isinstance(context,list):
            if context == 'block':
                top_level_elements = self.dialect.block_elements
            elif context == 'inline':
                top_level_elements = self.dialect.inline_elements
        else:
            top_level_elements = context

        if preprocess:
            text = self.preprocess(text)

        return bldr.tag(fragmentize(text,top_level_elements,element_store, environ))



    def generate(self,text,element_store=None,context='block', environ=None, preprocess=True):
        """Returns a Genshi Stream. See
        :meth:`~creoleparser.core.Parser.parse` for named parameter descriptions.

        """

        return self.parse(text, element_store, context, environ, preprocess).generate()


    def render(self, text, element_store=None, context='block', environ=None, preprocess=True, **kwargs):
        """Returns the final output string (e.g., xhtml). See
        :meth:`~creoleparser.core.Parser.parse` for named parameter descriptions.

        Left over keyword arguments (``kwargs``) will be passed to Genshi's Stream.render() method,
        overriding the corresponding attributes of the Parser object. For more infomation on Streams,
        see the `Genshi documentation <http://genshi.edgewall.org/wiki/Documentation/streams.html#serialization-options>`_.
        
        """

        kwargs.setdefault('method',self.method)
        kwargs.setdefault('encoding',self.encoding)
        if kwargs['method'] != "text":
            kwargs.setdefault('strip_whitespace',self.strip_whitespace)
        stream = self.generate(text, element_store, context, environ, preprocess)
        return stream.render(**kwargs)

    def __call__(self,text, **kwargs):
        """Wrapper for the render method. Returns final output string.

        """

        return self.render(text, **kwargs)

    def preprocess(self,text):
        """This should generally be called before fragmentize().

        :parameters:
          text
            text to be processsed.

        """
        text = text.replace("\r\n", "\n")
        text = text.replace("\r", "\n")

        return text    



class ArgParser(object):
    """Creates a callable object for parsing macro argument strings

    >>> from dialects import creepy20_base
    >>> my_parser = ArgParser(dialect=creepy20_base())
    >>> my_parser(" one two foo='three' boo='four' ")
    (['one', 'two'], {'foo': 'three', 'boo': 'four'})
 
    A parser returns a two-tuple, the first item being a list of positional
    arguments and the second a dictionary of keyword arguments. Argument
    values are either strings or lists.
    
    """
    
    def __init__(self,dialect, convert_implicit_lists=True,
                 arg_func=None, key_func=None, illegal_keys=(),
                 convert_unicode_keys=True):
        """Constructor for ArgParser objects

        :parameters:
          convert_unicode_keys
            If *True*, keys will be converted using ``str(key)`` before being
            added to the output dictionary. This allows the dictionary to be
            safely passed to functions using the special ``**`` form (i.e.,
            ``func(**kwargs)``).
          dialect
            Usually created using :func:`~creoleparser.dialects.creepy10_base`
            or :func:`~creoleparser.dialects.creepy20_base`
          convert_implicit_lists
            If *True*, all implicit lists will be converted to strings
            using ``' '.join(list)``. "Implicit" lists are created when
            positional arguments follow keyword arguments
            (see :func:`~creoleparser.dialects.creepy10_base`). 
          illegal_keys
            A tuple of keys that will be post-fixed with an underscore if found
            during parsing. 
          arg_func
            If supplied, this function will be used to transform the values
            of the positional arguments.
          key_func
            If supplied, this function will be used to transform the names
            and values of keyword arguments. It must accept two positional
            arguments. For example, this can be used to make keywords case
            insensitive:
            
            >>> from string import lower
            >>> from dialects import creepy20_base
            >>> my_parser = ArgParser(dialect=creepy20_base(),
            ...     key_func=lambda k, v: k, v.lower())
            >>> my_parser(" Foo='one' ")
            ([], {'foo': 'one'})
            
        """
    
        self.dialect = dialect()
        self.convert_implicit_lists = convert_implicit_lists
        self.arg_func = arg_func
        self.key_func = key_func
        self.illegal_keys = illegal_keys
        self.convert_unicode_keys = convert_unicode_keys


    def __call__(self, arg_string, **kwargs): 
        """Parses the ``arg_string`` returning a two-tuple 

        Keyword arguments (``kwargs``) can be used to override the corresponding
        attributes of the ArgParser object (see above). However, the
        ``dialect`` attribute **cannot** be overridden.
        
        """
        
        kwargs.setdefault('convert_implicit_lists',self.convert_implicit_lists)
        kwargs.setdefault('arg_func',self.arg_func)
        kwargs.setdefault('key_func',self.key_func)
        kwargs.setdefault('illegal_keys',self.illegal_keys)
        kwargs.setdefault('convert_unicode_keys',self.convert_unicode_keys)

        return self._parse(arg_string,**kwargs)


    def _parse(self,arg_string, convert_implicit_lists, arg_func, key_func,
            illegal_keys, convert_unicode_keys):
        
        frags = fragmentize(arg_string,self.dialect.top_elements,{},{},
                remove_escapes=False)
        positional_args = []
        kw_args = {}
        for arg in frags:
           if isinstance(arg,tuple):
             k, v  = arg
             if convert_unicode_keys:
                 k = str(k)
             if key_func:
                 k, v =  key_func(k, v)
             if k in illegal_keys:
                 k = k + '_'
             if k in kw_args:
                if isinstance(v,list):
                   try:
                      kw_args[k].extend(v)
                   except AttributeError:
                      v.insert(0,kw_args[k])
                      kw_args[k] = v
                elif isinstance(kw_args[k],list):
                   kw_args[k].append(v)
                else:
                   kw_args[k] = [kw_args[k], v]
                kw_args[k] = ImplicitList(kw_args[k])
             else:
                kw_args[k] = v
             if isinstance(kw_args[k],ImplicitList) and convert_implicit_lists:
                 kw_args[k] = ' ' .join(kw_args[k])
           else:
             if arg_func:
                 arg = arg_func(arg)
             positional_args.append(arg)

        return (positional_args, kw_args)
        



def fragmentize(text,wiki_elements, element_store, environ, remove_escapes=True):

    """Takes a string of wiki markup and outputs a list of genshi
    Fragments (Elements and strings).

    This recursive function, with help from the WikiElement objects,
    does almost all the parsing.

    When no WikiElement objects are supplied, escapes are removed from
    ``text`` (except if remove_escapes=True)  and it is
    returned as-is. This is the only way for recursion to stop.

    :parameters:
      text
        the text to be parsed
      wiki_elements
        list of WikiElement objects to be searched for
      environ
        object that may by used by macros
      remove_escapes
        If False, escapes will not be removed
    
    """

    while wiki_elements:
        # If the first supplied wiki_element is actually a list of elements, \
        # search for all of them and match the closest one only.
        if isinstance(wiki_elements[0],(list,tuple)):
            x = None
            mos = None
            for element in wiki_elements[0]:
                mo = element.regexp.search(text)
                if mo:
                    if x is None or mo.start() < x:
                        x,wiki_element,mos = mo.start(),element,[mo]
        else:
            wiki_element = wiki_elements[0]
            mos = [mo for mo in wiki_element.regexp.finditer(text)]
             
        if mos:
            frags = wiki_element._process(mos, text, wiki_elements, element_store, environ)
            break
        else:
            wiki_elements = wiki_elements[1:]

    # remove escape characters 
    else: 
        if remove_escapes:
            text = esc_to_remove.sub('',text)
        frags = fill_from_store(text,element_store)
        
    return frags


def fill_from_store(text,element_store):
    frags = []
    mos = place_holder_re.finditer(text)
    start = 0
    for mo in mos:
        if mo.start() > start:
            frags.append(text[start:mo.start()])
        frags.append(element_store.get(mo.group(1),
                       mo.group(1).join(['<<<','>>>'])))
        start = mo.end()
    if start < len(text):
        frags.append(text[start:])
    return frags


class ImplicitList(list):
    """This class marks argument lists as implicit"""
    pass

_pattern = r"""
    {(?:
      (?P<escaped>{) |   # Escape sequence of two delimiters
      (?P<named>[_a-z][_a-z0-9]*)}      |   # delimiter and a Python identifier
      (?P<braced>[_a-z][_a-z0-9]*)}   |   # delimiter and a braced identifier
      (?P<invalid>)              # Other ill-formed delimiter exprs
    )
    """

class P3Template(string.Template):
    pattern = _pattern
    delimiter = '{'

    def substitute(self, *args, **kws):
        return super(P3Template,self).substitute(*args, **kws).replace('}}','}')


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()




# Module:   dbm
# Date:     24th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""DataBase Manager

...
"""

from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from circuits import handler, BaseComponent, Event

import schema

metadata = MetaData()
Base = declarative_base(metadata=metadata)

class DatabaseLoaded(Event):
    """Database Loaded Event"""

class DatabaseManager(BaseComponent):

    def __init__(self, dburi, echo=False, convert_unicode=True):
        super(DatabaseManager, self).__init__()

        self.engine = create_engine(dburi,
            echo=echo,
            convert_unicode=convert_unicode,
        )

        self.session = scoped_session(
            sessionmaker(
                bind=self.engine,
                autoflush=True,
                autocommit=False,
            )
        )

    @handler("registered")
    def _on_registered(self, component, manager):
        if component == self:
            tables = self.engine.table_names()
            for Table, data in schema.DATA:
                if Table.__tablename__ not in tables:
                    Table.__table__.create(self.engine)
                    for row in data:
                        self.session.add(Table(*row))
                    self.session.commit()
            metadata.create_all(self.engine)

    @handler("stopped", target="*")
    def _on_stopped(self, component):
        self.session.flush()
        self.session.commit()
        self.session.close()



# Module:   errors
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Error Definitions

...
"""

from circuits.web.exceptions import HTTPException

class WikiError(HTTPException):
    """Base class for all error pages."""

    traceback = False

class ForbiddenErr(WikiError):

    code = 403

class NotFoundErr(WikiError):

    code = 404

class UnsupportedMediaTypeErr(WikiError):

    code = 415

class NotImplementedErr(WikiError):

    code = 501

    description = (
        "<p>This feature has not yet been implemented."
        "Please come back and try again later.</p>"
    )

class ServiceUnavailableErr(WikiError):

    code = 503



# Module:   config
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Configuration Handling

...
"""

import os
import ConfigParser
from warnings import warn

from argparse import ArgumentParser

import reprconf
import sahriswiki

class Config(reprconf.Config):

    prefix = "SAHRIS_"

    def __init__(self, file=None, **kwargs):
        super(Config, self).__init__(file, **kwargs)

        self.parse_environ()
        self.parse_options()
        self.check_options()

    def check_options(self):
        paths = ("accesslog", "config", "repo", "errorlog",
                "pidfile", "sock", "theme",)

        for path in paths:
            value = self.get(path, None)
            if value and not os.path.isabs(value):
                self[path] = os.path.abspath(os.path.expanduser(value))

    def parse_environ(self):
        """Check the environment variables for options."""

        config = {}

        for key, value in os.environ.iteritems():
            if key.startswith(self.prefix):
                name = key[len(self.prefix):].lower()
                config[name] = value

        self.update(config)

    def parse_options(self):
        parser = ArgumentParser(sahriswiki.__name__)

        add = parser.add_argument

        add("--config", action="store", default=None,
                dest="config", metavar="FILE", type=str,
                help="Read configuration from FILE")

        add("-b", "--bind", action="store", default="0.0.0.0",
                dest="bind", metavar="INT", type=str,
                help="Listen on interface INT")

        add("-p", "--port", action="store", default=8000,
                dest="port", metavar="PORT", type=int,
                help="Listen on port PORT")

        add("-s", "--sock", action="store", default=None,
                dest="sock", metavar="FILE", type=str,
                help="Listen on socket FILE")

        add("-r", "--repo", action="store", default="wiki",
                dest="repo", metavar="REPO", type=str,
                help="Store pages in mercurial repository REPO")

        add("-d", "--database", action="store",
                default="sqlite:///sahriswiki.db",
                dest="db", metavar="DB", type=str,
                help="Store meta data in database DB")

        add("-t", "--theme", action="store",
                default=os.path.join(os.path.dirname(__file__)),
                dest="theme", metavar="DIR", type=str,
                help="Set theme (static and templates) path to DIR")

        add("--frontpage", action="store", default="FrontPage",
                dest="frontpage", metavar="PAGE", type=str,
                help="Set default front page to PAGE")

        add("--index", action="store", default="Index",
                dest="index", metavar="PAGE", type=str,
                help="Set default index page to PAGE")

        add("--indexes", action="store", nargs="+",
            default=["Index", "index.html", "index.rst"],
            dest="indexes", metavar="LIST", type=list,
            help="Set index search list to LIST")

        add("--menu", action="store", default="SiteMenu",
                dest="menu", metavar="PAGE", type=str,
                help="Set default site menu page to PAGE")

        add("--encoding", action="store", default="utf-8",
                dest="encoding", metavar="ENC", type=str,
                help="Use encoding ENC to read and write pages")

        add("--language", action="store", default="en",
                dest="language", metavar="LANG", type=str,
                help="Translate interface to LANG")

        add("--name", action="store", default=sahriswiki.__name__,
                dest="name", metavar="NAME", type=str,
                help="Set site name to NAME")

        add("--author", action="store", default=sahriswiki.__author__,
                dest="author", metavar="NAME", type=str,
                help="Set site author to NAME")

        add("--description", action="store",
                default=sahriswiki.__doc__.split("\n")[0],
                dest="description", metavar="DESC", type=str,
                help="Set site description to DESC")

        add("--keywords", action="store", default="sahriswiki",
                dest="keywords", metavar="KEYWORDS", type=str,
                help="Set site keywords to KEYWORDS")

        add("--readonly", action="store_true", default=False,
                dest="readonly",
                help="Set the wiki into readonly mode")

        add("--debug", action="store_true", default=False,
                dest="debug",
                help="Enable debugging mode")

        add("--daemon", action="store_true", default=False,
                dest="daemon",
                help="Run as a background process")

        add("--verbose", action="store_true", default=False,
                dest="verbose",
                help="Enable verbose debugging")

        add("--errorlog", action="store", default=None,
                dest="errorlog", metavar="FILE", type=str,
                help="Store debug and error information in FILE")

        add("--accesslog", action="store", default=None,
                dest="accesslog", metavar="FILE", type=str,
                help="Store web server access logs in FILE")

        add("--pidfile", action="store", default="sahris.pid",
                dest="pidfile", metavar="FILE", type=str,
                help="Write process id to FILE")

        add("--disable-hgweb", action="store_true", default=False,
                dest="disable-hgweb",
                help="Disable hgweb interface")

        add("--disable-logging", action="store_true", default=False,
                dest="disable-logging",
                help="Disable access logging")

        add("--disable-static", action="store_true", default=False,
                dest="disable-static",
                help="Disable static file serving")

        add("--disable-compression", action="store_true", default=False,
                dest="disable-compression",
                help="Disable compression")

        add("--static-baseurl", action="store", default=None,
                dest="static-baseurl", metavar="URL", type=str,
                help="Set static baseurl to URL")

        namespace = parser.parse_args()

        if namespace.config is not None:
            filename = os.path.abspath(os.path.expanduser(namespace.config))
            if os.path.exists(filename) and os.path.isfile(filename):
                config = reprconf.as_dict(filename)
                if config.keys():
                    config = config[config.keys()[0]]
                    for option, value in config.iteritems():
                        if option in namespace:
                            self[option] = value
                        else:
                            warn("Ignoring unknown option %r" % option)

        for option, value in namespace.__dict__.iteritems():
            if option not in self and value is not None:
                self[option] = value

    def reload_config(self):
        filename = self.get("config")
        if filename is not None:
            config = reprconf.as_dict(filename)
            self.update(config.get(config.keys()[0], {}))

    def save_config(self, filename=None):
        if filename is None:
            filename = self.get("config", "sahris.conf")

        parser = ConfigParser.RawConfigParser()
        section = self.get("name", "sahris")
        parser.add_section(section)
        for key, value in self.iteritems():
            parser.set(section, str(key), repr(value))

        configfile = open(filename, "wb")
        try:
            parser.write(configfile)
        finally:
            configfile.close()



# Module:   reprconf
# Date:     15th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au
# NOTE:     Borrowed from CherryPy3

"""Generic configuration system using unrepr.

Configuration data may be supplied as a Python dictionary, as a filename,
or as an open file object. When you supply a filename or file, Python's
builtin ConfigParser is used (with some extensions).
"""

from ConfigParser import ConfigParser

from unrepr import unrepr

def as_dict(config):
    """Return a dict from 'config' whether it is a dict, file, or filename."""
    if isinstance(config, str):
        config = Parser().dict_from_file(config)
    elif hasattr(config, 'read'):
        config = Parser().dict_from_file(config)
    return config

class Config(dict):
    """A dict-like set of configuration data.
    
    May take a file, filename, or dict.
    """
    
    defaults = {}
    environments = {}
    
    def __init__(self, file=None, **kwargs):
        super(Config, self).__init__()

        self.reset()
        if file is not None:
            self.update(file)
        if kwargs:
            self.update(kwargs)
    
    def reset(self):
        """Reset self to default values."""
        self.clear()
        dict.update(self, self.defaults)
    
    def update(self, config):
        """Update self from a dict, file or filename."""
        if isinstance(config, str):
            # Filename
            config = Parser().dict_from_file(config)
        elif hasattr(config, 'read'):
            # Open file object
            config = Parser().dict_from_file(config)
        else:
            config = config.copy()
        self._apply(config)
    
    def _apply(self, config):
        """Update self from a dict."""
        which_env = config.get('environment')
        if which_env:
            env = self.environments[which_env]
            for k in env:
                if k not in config:
                    config[k] = env[k]
        
        dict.update(self, config)

    def __setitem__(self, k, v):
        dict.__setitem__(self, k, v)

class Parser(ConfigParser):
    """Sub-class of ConfigParser that keeps the case of options and that raises
    an exception if the file cannot be read.
    """
    
    def optionxform(self, optionstr):
        return optionstr
    
    def read(self, filenames):
        if isinstance(filenames, str):
            filenames = [filenames]
        for filename in filenames:
            # try:
            #     fp = open(filename)
            # except IOError:
            #     continue
            fp = open(filename)
            try:
                self._read(fp, filename)
            finally:
                fp.close()
    
    def as_dict(self, raw=False, vars=None):
        """Convert an INI file to a dictionary"""
        # Load INI file into a dict
        result = {}
        for section in self.sections():
            if section not in result:
                result[section] = {}
            for option in self.options(section):
                value = self.get(section, option, raw, vars)
                try:
                    value = unrepr(value)
                except Exception as x:
                    msg = ("Config error in section: %r, option: %r, "
                           "value: %r. Config values must be valid Python." %
                           (section, option, value))
                    raise ValueError(msg, x.__class__.__name__, x.args)
                result[section][option] = value
        return result
    
    def dict_from_file(self, file):
        if hasattr(file, 'read'):
            self.readfp(file)
        else:
            self.read(file)
        return self.as_dict()



# -*- coding: utf-8 -*-
# Module:   utils
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Utility Functions

...
"""

import re

from circuits.web.tools import mimetypes

NEWLINES = re.compile("\n|\r[^\n]|\r\n")

def external_link(addr):
    """
    Decide whether a link is absolute or internal.

    >>> external_link('http://example.com')
    True
    >>> external_link('https://example.com')
    True
    >>> external_link('ftp://example.com')
    True
    >>> external_link('mailto:user@example.com')
    True
    >>> external_link('PageTitle')
    False
    >>> external_link(u'ąęśćUnicodePage')
    False

    """

    return (addr.startswith('http://')
            or addr.startswith('https://')
            or addr.startswith('ftp://')
            or addr.startswith('mailto:'))

def page_mime(title, types=[("+", "type")], default="text/x-wiki"):
    """
    Guess page's mime type ased on corresponding file name.

    Files that start with a type prefix (the first item of each item
    in the type list) then the type if returned for that file (the
    second item of th matching item in the type list).

    If nothing else matches, default is returned.

    >>> page_mime(u'something.txt')
    'text/plain'
    >>> page_mime(u'SomePage')
    'text/x-wiki'
    >>> page_mime(u'ąęśUnicodePage')
    'text/x-wiki'
    >>> page_mime(u'image.png')
    'image/png'
    >>> page_mime(u'style.css')
    'text/css'
    >>> page_mime(u'archive.tar.gz')
    'archive/gzip'
    >>> page_mime(u'+history')
    'type/history'
    """

    addr = title.encode('utf-8') # the encoding doesn't relly matter here
    mime, encoding = mimetypes.guess_type(addr, strict=False)

    if encoding:
        return 'archive/%s' % encoding

    if mime is None:
        for prefix, type in types:
            if title.startswith(prefix):
                return "/".join([type, title[title.find(prefix) + 1:]])
        return default
    return mime

def extract_links(text):
    links = re.compile(ur"\[\[(?P<link_target>([^|\]]|\][^|\]])+)"
            ur"(\|(?P<link_text>([^\]]|\][^\]])+))?\]\]")

    for m in links.finditer(text):
        if m.groupdict():
            d = m.groupdict()
            yield d["link_target"], d["link_text"] or ""



# Module:   schema
# Date:     25th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""(Default) Schema(s) / Data

...
"""

from hashlib import md5

from sqlalchemy import Boolean, Column, Integer, PickleType, Sequence, String

from dbm import Base

version = 1

class System(Base):

    __tablename__ = "system"

    name = Column(String(20), primary_key=True)
    value = Column(String(80))

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "<System('%s', '%s')>" % (self.name, self.value)

class User(Base):

    __tablename__ = "users"

    username = Column(String(20), primary_key=True)
    password = Column(String(256))

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def __repr__(self):
        return "<User('%s')>" % self.username

class Permission(Base):

    __tablename__ = "permissions"

    username = Column(String(20), primary_key=True)
    action = Column(String(20), primary_key=True)

    def __init__(self, username, action):
        self.username = username
        self.action = action

    def __repr__(self):
        return "<Permission('%s', '%s')>" % (self.username, self.action)

class Session(Base):

    __tablename__ = "sessions"

    sid = Column(String(256), primary_key=True)
    authenticated = Column(Boolean, index=True, primary_key=True)
    data = Column(PickleType)
    time = Column(Integer, index=True)

    def __init__(self, sid, authenticated):
        self.sid = sid
        self.authenticated = authenticated

    def __repr__(self):
        return "<Session('%s', '%s')>" % (self.sid, self.authenticated)

DATA = (
    (User, (
        ("admin", md5("admin").hexdigest()),
    )),
    (Permission, (
        ("anonymous",     "PAGE_DOWNLOAD"),
        ("anonymous",     "PAGE_HISTORY"),
        ("anonymous",     "PAGE_VIEW"),
        ("anonymous",     "SEARCH_VIEW"),
        ("authenticated", "PAGE_EDIT"),
        ("authenticated", "PAGE_DELETE"),
        ("authenticated", "PAGE_RENAME"),
        ("authenticated", "PAGE_UPLOAD"),
        ("admin",         "SAHRIS_ADMIN"),
    )),
    (System, (
        ("schema_version", str(version)),
    )),
)



# Module:   storage
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Storage Classes

...
"""

import re
import os
import thread
import tempfile

import mercurial.hg
import mercurial.ui
import mercurial.util
import mercurial.revlog
import mercurial.context
from mercurial.node import short

from circuits.web.utils import url_quote, url_unquote

from i18n import _
from errors import ForbiddenErr, NotFoundErr

def locked_repo(func):
    """A decorator for locking the repository when calling a method."""

    def new_func(self, *args, **kwargs):
        """Wrap the original function in locks."""

        wlock = self.repo.wlock()
        lock = self.repo.lock()
        try:
            func(self, *args, **kwargs)
        finally:
            lock.release()
            wlock.release()

    return new_func

class WikiStorage(object):
    """
    Provides means of storing wiki pages and keeping track of their
    change history, using Mercurial repository as the storage method.
    """

    def __init__(self, path, charset=None):
        """
        Takes the path to the directory where the pages are to be kept.
        If the directory doen't exist, it will be created. If it's inside
        a Mercurial repository, that repository will be used, otherwise
        a new repository will be created in it.
        """

        self.charset = charset or 'utf-8'
        self.path = os.path.abspath(path)
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        self.repo_path = self._find_repo_path(self.path)
        try:
            self.ui = mercurial.ui.ui(report_untrusted=False,
                                      interactive=False, quiet=True)
        except TypeError:
            # Mercurial 1.3 changed the way we setup the ui object.
            self.ui = mercurial.ui.ui()
            self.ui.quiet = True
            self.ui._report_untrusted = False
            self.ui.setconfig('ui', 'interactive', False)
        if self.repo_path is None:
            self.repo_path = self.path
            create = True
        else:
            create = False
        self.repo_prefix = self.path[len(self.repo_path):].strip('/')
        self._repos = {}
        self._workingctxs = {}
        # Create the repository if needed.
        mercurial.hg.repository(self.ui, self.repo_path, create=create)

    def reopen(self):
        """Close and reopen the repo, to make sure we are up to date."""

        #self.repo = mercurial.hg.repository(self.ui, self.repo_path)
        self._repos = {}
        self._workingctxs = {}

    @property
    def repo(self):
        """Keep one open repository per thread."""

        thread_id = thread.get_ident()
        try:
            return self._repos[thread_id]
        except KeyError:
            repo = mercurial.hg.repository(self.ui, self.repo_path)
            self._repos[thread_id] = repo
            return repo

    @property
    def workingctx(self):
        """Keep one open working ctx per thread."""

        thread_id = thread.get_ident()
        try:
            return self._workingctxs[thread_id]
        except KeyError:
            workingctx = mercurial.context.workingctx(self.repo)
            self._workingctxs[thread_id] = workingctx
            return workingctx

    def _find_repo_path(self, path):
        """Go up the directory tree looking for a repository."""

        while not os.path.isdir(os.path.join(path, ".hg")):
            old_path, path = path, os.path.dirname(path)
            if path == old_path:
                return None
        return path

    def _check_path(self, path):
        """
        Ensure that the path is within allowed bounds.
        """

        abspath = os.path.abspath(path)
        if os.path.islink(path) or os.path.isdir(path):
            raise ForbiddenErr(
                _(u"Can't use symbolic links or directories as pages"))
        if not abspath.startswith(self.path):
            raise ForbiddenErr(
                _(u"Can't read or write outside of the pages repository"))

    def _file_path(self, title):
        return os.path.join(self.repo_path, self._title_to_file(title))

    def _title_to_file(self, title):
        title = unicode(title).strip()
        filename = url_quote(title, safe='')
        # Escape special windows filenames and dot files
        _windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3',
                                 'COM4', 'LPT1', 'LPT2', 'LPT3', 'PRN',
                                 'NUL')
        if (filename.split('.')[0].upper() in _windows_device_files or
            filename.startswith('_') or filename.startswith('.')):
            filename = '_' + filename
        return os.path.join(self.repo_prefix, filename)

    def _file_to_title(self, filepath):
        if not filepath.startswith(self.repo_prefix):
            raise ForbiddenErr(
                _(u"Can't read or write outside of the pages repository"))
        name = filepath[len(self.repo_prefix):].strip('/')
        # Unescape special windows filenames and dot files
        if name.startswith('_') and len(name)>1:
            name = name[1:]
        return url_unquote(name)

    def __contains__(self, title):
        if title:
            file_path = self._file_path(title)
            return os.path.isfile(file_path) and not os.path.islink(file_path)

    def __iter__(self):
        return self.all_pages()

    def merge_changes(self, changectx, repo_file, text, user, parent):
        """Commits and merges conflicting changes in the repository."""

        tip_node = changectx.node()
        # FIXME: The following line fails sometimes :/
        filectx = changectx[repo_file].filectx(parent)
        parent_node = filectx.changectx().node()

        self.repo.dirstate.setparents(parent_node)
        node = self._commit([repo_file], text, user)

        partial = lambda filename: repo_file == filename
        try:
            mercurial.merge.update(self.repo, tip_node, True, True, partial)
            msg = _(u'merge of edit conflict')
        except mercurial.util.Abort:
            msg = _(u'failed merge of edit conflict')
        self.repo.dirstate.setparents(tip_node, node)
        # Mercurial 1.1 and later need updating the merge state
        try:
            mercurial.merge.mergestate(self.repo).mark(repo_file, "r")
        except (AttributeError, KeyError):
            pass
        return msg

    @locked_repo
    def save_file(self, title, file_name, author, comment, parent=None):
        """Save an existing file as specified page."""

        user = author.encode('utf-8')
        text = comment.encode('utf-8')
        repo_file = self._title_to_file(title)
        file_path = self._file_path(title)
        self._check_path(file_path)
        mercurial.util.rename(file_name, file_path)
        changectx = self._changectx()
        try:
            filectx_tip = changectx[repo_file]
            current_page_ver = (
                filectx_tip.filerev(),
                short(filectx_tip.node())
            )
        except mercurial.revlog.LookupError:
            self.workingctx.add([repo_file])
            current_page_ver = []
        if parent is not None and parent not in current_page_ver:
            msg = self.merge_changes(changectx, repo_file, text, user, parent)
            user = '<wiki>'
            text = msg.encode('utf-8')
        self._commit([repo_file], text, user)


    def _commit(self, files, text, user):
        try:
            return self.repo.commit(files=files, text=text, user=user,
                                    force=True, empty_ok=True)
        except TypeError:
            # Mercurial 1.3 doesn't accept empty_ok or files parameter
            match = mercurial.match.exact(self.repo_path, '', list(files))
            return self.repo.commit(match=match, text=text, user=user,
                                    force=True)


    def save_data(self, title, data, author=u'', comment=u'', parent=None):
        """Save data as specified page."""

        try:
            temp_path = tempfile.mkdtemp(dir=self.path)
            file_path = os.path.join(temp_path, 'saved')
            f = open(file_path, "wb")
            f.write(data)
            f.close()
            self.save_file(title, file_path, author, comment, parent)
        finally:
            try:
                os.unlink(file_path)
            except OSError:
                pass
            try:
                os.rmdir(temp_path)
            except OSError:
                pass

    def save_text(self, title, text, author=u'', comment=u'', parent=None):
        """Save text as specified page, encoded to charset."""

        data = text.encode(self.charset)
        self.save_data(title, data, author, comment, parent)

    def page_text(self, title):
        """Read unicode text of a page."""

        data = self.open_page(title).read()
        text = unicode(data, self.charset, 'replace')
        return text

    def page_lines(self, page):
        for data in page.xreadlines():
            yield unicode(data, self.charset, 'replace')

    @locked_repo
    def delete_page(self, title, author=u'', comment=u''):
        user = author.encode('utf-8') or 'anon'
        text = comment.encode('utf-8') or 'deleted'
        repo_file = self._title_to_file(title)
        file_path = self._file_path(title)
        self._check_path(file_path)
        try:
            os.unlink(file_path)
        except OSError:
            pass
        self.workingctx.remove([repo_file])
        self._commit([repo_file], text, user)

    def open_page(self, title):
        """Open the page and return a file-like object with its contents."""

        file_path = self._file_path(title)
        self._check_path(file_path)
        try:
            return open(file_path, "rb")
        except IOError:
            raise NotFoundErr()

    def page_file_meta(self, title):
        """Get page's inode number, size and last modification time."""

        try:
            (st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size,
             st_atime, st_mtime, st_ctime) = os.stat(self._file_path(title))
        except OSError:
            return 0, 0, 0
        return st_ino, st_size, st_mtime

    def page_meta(self, title):
        """Get page's revision, date, last editor and his edit comment."""

        filectx_tip = self._find_filectx(title)
        if filectx_tip is None:
            raise NotFoundErr()
            #return -1, None, u'', u''
        rev = filectx_tip.filerev()
        node = filectx_tip.node()
        filectx = filectx_tip.filectx(rev)
        date = filectx.date()[0]
        author = unicode(filectx.user(), "utf-8",
                         'replace').split('<')[0].strip()
        comment = unicode(filectx.description(), "utf-8", 'replace')
        return rev, node, date, author, comment

    def repo_revision(self):
        """Give the latest revision of the repository."""

        return self._changectx().rev()

    def repo_node(self):
        """Give the latest node of the repository."""

        return self._changectx().node()

    def _changectx(self):
        """Get the changectx of the tip."""

        try:
            # This is for Mercurial 1.0
            return self.repo.changectx()
        except TypeError:
            # Mercurial 1.3 (and possibly earlier) needs an argument
            return self.repo.changectx('tip')

    def _find_filectx(self, title):
        """Find the last revision in which the file existed."""

        stack = [self._changectx()]
        repo_file = self._title_to_file(title)

        while stack:
            changectx = stack.pop()
            if changectx.rev() == 0:
                return None
            if repo_file in changectx:
               return changectx[repo_file]
            else:
                for parent in changectx.parents():
                    if parent and parent not in stack:
                        stack.append(parent)

    def page_history(self, title):
        """Iterate over the page's history."""

        filectx_tip = self._find_filectx(title)
        if filectx_tip is None:
            return
        maxrev = filectx_tip.filerev()
        minrev = 0
        for rev in range(maxrev, minrev-1, -1):
            filectx = filectx_tip.filectx(rev)
            date = filectx.date()[0]
            author = unicode(filectx.user(), "utf-8",
                             'replace').split('<')[0].strip()
            comment = unicode(filectx.description(), "utf-8", 'replace')
            yield rev, date, author, comment

    def page_revision(self, title, rev):
        """Get binary content of the specified revision of the page."""

        filectx_tip = self._find_filectx(title)
        if filectx_tip is None:
            raise NotFoundErr()
        try:
            data = filectx_tip.filectx(rev).data()
        except IndexError:
            raise NotFoundErr()
        return data

    def revision_text(self, title, rev):
        """Get unicode text of the specified revision of the page."""

        data = self.page_revision(title, rev)
        text = unicode(data, self.charset, 'replace')
        return text

    def history(self):
        """Iterate over the history of entire wiki."""

        changectx = self._changectx()
        maxrev = changectx.rev()
        minrev = 0
        for wiki_rev in range(maxrev, minrev-1, -1):
            change = self.repo.changectx(wiki_rev)
            date = change.date()[0]
            author = unicode(change.user(), "utf-8",
                             'replace').split('<')[0].strip()
            comment = unicode(change.description(), "utf-8", 'replace')
            for repo_file in change.files():
                if repo_file.startswith(self.repo_prefix):
                    title = self._file_to_title(repo_file)
                    try:
                        rev = change[repo_file].filerev()
                    except mercurial.revlog.LookupError:
                        rev = -1
                    yield title, rev, date, author, comment

    def all_pages(self):
        """Iterate over the titles of all pages in the wiki."""

        for filename in os.listdir(self.path):
            file_path = os.path.join(self.path, filename)
            if (os.path.isfile(file_path)
                and not os.path.islink(file_path)
                and not filename.startswith('.')):
                yield url_unquote(filename)

    def changed_since(self, rev):
        """Return all pages that changed since specified repository revision."""

        try:
            last = self.repo.lookup(int(rev))
        except IndexError:
            for page in self.all_pages():
                yield page
            return
        current = self.repo.lookup('tip')
        status = self.repo.status(current, last)
        modified, added, removed, deleted, unknown, ignored, clean = status
        for filename in modified+added+removed+deleted:
            if filename.startswith(self.repo_prefix):
                yield self._file_to_title(filename)


class WikiSubdirectoryStorage(WikiStorage):
    """
    A version of WikiStorage that keeps the subpages in real subdirectories in
    the filesystem.

    """

    periods_re = re.compile(r'^[.]|(?<=/)[.]')
    slashes_re = re.compile(r'^[/]|(?<=/)[/]')

    def _title_to_file(self, title):
        """Modified escaping allowing (some) slashes and spaces."""

        title = unicode(title).strip()
        escaped = url_quote(title, safe='/ ')
        escaped = self.periods_re.sub('%2E', escaped)
        escaped = self.slashes_re.sub('%2F', escaped)
        path = os.path.join(self.repo_prefix, escaped)
        return path

    @locked_repo
    def save_file(self, title, file_name, author=u'', comment=u'', parent=None):
        """
        Save the file and make the subdirectories if needed.
        """

        file_path = self._file_path(title)
        self._check_path(file_path)
        dir_path = os.path.dirname(file_path)
        try:
            os.makedirs(dir_path)
        except OSError, e:
            if e.errno == 17 and not os.path.isdir(dir_path):
                raise ForbiddenErr(
                    _(u"Can't make subpages of existing pages"))
            elif e.errno != 17:
                raise
        super(WikiSubdirectoryStorage, self).save_file(title, file_name,
                                                       author, comment, parent)

    @locked_repo
    def delete_page(self, title, author=u'', comment=u''):
        """
        Remove empty directories after deleting a page.

        Note that Mercurial doesn't track directories, so we don't have to
        commit after removing empty directories.
        """

        super(WikiSubdirectoryStorage, self).delete_page(title, author, comment)
        file_path = self._file_path(title)
        self._check_path(file_path)
        dir_path = os.path.dirname(file_path)
        try:
            os.removedirs(dir_path)
        except OSError, e:
            pass # Ignore possibly OSError (39) Directory not empty errors.

    def all_pages(self):
        """
        Iterate over the titles of all pages in the wiki.
        Include subdirectories.
        """

        for (dirpath, dirnames, filenames) in os.walk(self.path):
            path = dirpath[len(self.path)+1:]
            for name in filenames:
                filename = os.path.join(path, name)
                if (os.path.isfile(os.path.join(self.path, filename))
                    and not filename.startswith('.')):
                    yield url_unquote(filename)

    def all_pages_tree(self):
        """
        Iterate over the titles of all pages in the wiki.
        Include subdirectories.

        Return a tree of al pages.
        """

        def generate(root):
            for name in os.listdir(root):
                if name.startswith("."):
                    continue
                path = os.path.join(root, name)
                if os.path.isdir(path):
                    yield {name: sorted(generate(path))}
                elif os.path.isfile(path) and not name.startswith("."):
                    rel = os.path.relpath(path, self.path)
                    yield url_unquote(rel),  url_unquote(name)

        return generate(self.path)

class WikiSubdirectoryIndexesStorage(WikiSubdirectoryStorage):
    """
    A version of WikiSubdirectoryStorage that defaults to a set of indexes.
    """

    index = "Index" # Default index
    indexes = ["FrontPage", "Index"] # Default list of search indexes

    def __init__(self, path, charset=None, **kwargs):
        super(WikiSubdirectoryIndexesStorage, self).__init__(path, charset)

        if "index" in kwargs:
            self.index = kwargs["index"]

        if "indexes" in kwargs:
            self.indexes = kwargs["indexes"]

    def _file_path(self, title):
        root = super(WikiSubdirectoryIndexesStorage, self)._file_path(title)

        if os.path.isfile(root) and not os.path.islink(root):
            return root
        elif os.path.isdir(root):
            for index in self.indexes:
                path = os.path.join(root, index)
                if os.path.isfile(path) and not os.path.islink(path):
                    return path
            return os.path.join(root, self.indexes[0])
        return root

    def _title_to_file(self, title):
        root = super(WikiSubdirectoryIndexesStorage, self)._title_to_file(title)

        def exists(path):
            file_path = os.path.join(self.repo_path, path)
            return os.path.isfile(file_path) and not os.path.islink(file_path)

        def isdir(path):
            file_path = os.path.join(self.repo_path, path)
            return os.path.isdir(file_path) and not os.path.islink(file_path)

        if not exists(root):
            for index in self.indexes:
                path = os.path.join(root, index)
                if exists(path):
                    return path
            if isdir(root):
                return os.path.join(root, self.index)

        return root

    def all_pages(self):
        """
        Iterate over the titles of all pages in the wiki.
        Include subdirectories but skip over indexes.
        """

        for (dirpath, dirnames, filenames) in os.walk(self.path):
            path = dirpath[len(self.path)+1:]
            for name in filenames:
                if os.path.basename(name) in self.indexes:
                    filename = os.path.join(path, os.path.dirname(name))
                    yield url_unquote(filename)
                else:
                    filename = os.path.join(path, name)
                    if (os.path.isfile(os.path.join(self.path, filename))
                        and not filename.startswith('.')):
                        yield url_unquote(filename)

    def all_pages_tree(self):
        """
        Iterate over the titles of all pages in the wiki.
        Include subdirectories but skip over indexes.

        Return a tree of al pages.
        """

        def generate(root):
            for name in os.listdir(root):
                if name.startswith("."):
                    continue
                path = os.path.join(root, name)
                rel = os.path.relpath(path, self.path)
                base = os.path.dirname(rel)
                if os.path.isdir(path):
                    has_index = any([os.path.join(base, name, index) in self
                        for index in self.indexes])
                    yield {(url_unquote(rel), url_unquote(name), has_index):
                            sorted(generate(path))}
                elif os.path.isfile(path) and not os.path.islink(path) and \
                        name not in self.indexes:
                    yield url_unquote(rel),  url_unquote(name)

        return generate(self.path)

    def page_parent(self, title):
        filename = self._file_path(title)
        parent = os.path.dirname(filename)

        if os.path.isdir(parent) and not os.path.islink(parent):
            type = "dir"
        elif os.path.isfile(parent) and not os.path.islink(parent):
            type = "file"
        else:
            parent = None
            type = None

        if parent is not None:
            parent = os.path.relpath(parent, self.path)

        return parent, type

    @locked_repo
    def save_file(self, title, file_name, author=u'', comment=u'', parent=None):
        """
        Save the file and make the subdirectories if needed.
        """

        super(WikiSubdirectoryIndexesStorage, self).save_file(
                title, file_name, author, comment, parent)



# Module:   macros
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Macro Support

Macro loader and dispatcher for creoleparser
"""

import os
from traceback import format_exc
from inspect import getmembers, getmodule, isfunction

from genshi.builder import tag

from sahriswiki.unrepr import unrepr
from sahriswiki.highlight import highlight
from sahriswiki.creoleparser.core import ArgParser
from sahriswiki.creoleparser.dialects import ArgDialect
from sahriswiki.creoleparser.elements import KeywordArg, WhiteSpace

def creepy10_base():

    class Base(ArgDialect):

        kw_arg = KeywordArg(token='=')
        spaces = WhiteSpace()

        def __init__(self):
            self.kw_arg.child_elements = [self.spaces]

        @property
        def top_elements(self):
            return [self.kw_arg]

    return Base

def arg_func(arg):
    try:
        return unrepr(arg)
    except:
        return arg

def key_func(k, v):
    try:
        return k, unrepr(v)
    except:
        return k, v

parse_args = ArgParser(
    dialect=creepy10_base(),
    arg_func=arg_func,
    key_func=key_func,
)

class Macro(object):

    def __init__(self, name, arg_string, body, isblock):
        super(Macro, self).__init__()

        self.name = name
        self.arg_string = arg_string
        self.body = body
        self.isblock = isblock

def dispatcher(name, arg_string, body, isblock, (environ, data)):
    if name in environ.macros:
        macro = Macro(name, arg_string, body, isblock)
        args, kwargs = parse_args(arg_string)
        try:
            return environ.macros[name](macro, environ, data,
                *args, **kwargs)
        except Exception, e:
            error = "ERROR: Error while executing macro %s (%s)" % (name, e)
            traceback = format_exc()
            return tag.div(
                tag.p(error),
                highlight(traceback, lang="pytb"),
                class_="error"
            )

    else:
        return tag.div(tag.p("Macro %s Not Found!" % name), class_="error")

def loadMacros():
    path = os.path.abspath(os.path.dirname(__file__))
    p = lambda x: os.path.splitext(x)[1] == ".py"
    modules = [x for x in os.listdir(path) if p(x) and not x == "__init__.py"]

    macros = {}

    for module in modules:
        name, _ = os.path.splitext(module)

        moduleName = "%s.%s" % (__package__, name)
        m = __import__(moduleName, globals(), locals(), __package__)

        p = lambda x: isfunction(x) and getmodule(x) is m
        for name, function in getmembers(m, p):
            name = name.replace("_", "-")
            try:
                macros[name] = function
            except Exception, e:
                continue

    return macros



# Feedformatter
# Copyright (c) 2008, Luke Maurits <luke@maurits.id.au>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__version__ = "TRUNK"

from cStringIO import StringIO

# This "staircase" of import attempts is ugly.  If there's a nicer way to do
# this, please let me know!
try:
    import xml.etree.cElementTree as ET
except ImportError:
    try:
        import xml.etree.ElementTree as ET
    except ImportError:
        try:
            import cElementTree as ET
        except ImportError:
            try:
                from elementtree import ElementTree as ET
            except ImportError:
                raise ImportError("Could not import any form of element tree!")

try:
    from xml.dom.ext import PrettyPrint
    from xml.dom.ext.reader.Sax import FromXml
    feedformatterCanPrettyPrint = True
except ImportError:
    feedformatterCanPrettyPrint = False

from time import strftime, strptime, localtime, struct_time, timezone

# RSS 1.0 Functions ----------

_rss1_channel_mappings = (
    (("title",), "title"),
    (("link", "url"), "link"),
    (("description", "desc", "summary"), "description")
)

_rss1_item_mappings = (
    (("title",), "title"),
    (("link", "url"), "link"),
    (("description", "desc", "summary"), "description")
)

# RSS 2.0 Functions ----------

_rss2_channel_mappings = (
    (("title",), "title"),
    (("link", "url"), "link"),
    (("description", "desc", "summary"), "description"),
    (("pubDate", "pubdate", "date", "published", "updated"), "pubDate", lambda(x): _format_datetime("rss2",x)),
    (("category",), "category"),
    (("language",), "language"),
    (("copyright",), "copyright"),
    (("webMaster",), "webmaster"),
    (("image",), "image"),
    (("skipHours",), "skipHours"),
    (("skipDays",), "skipDays")
)

_rss2_item_mappings = (
    (("title",), "title"),
    (("link", "url"), "link"),
    (("description", "desc", "summary"), "description"),
    (("guid", "id"), "guid"),
    (("pubDate", "pubdate", "date", "published", "updated"), "pubDate", lambda(x): _format_datetime("rss2",x)),
    (("category",), "category"),
    (("author",), "author", lambda(x): _rssify_author(x))
)

# Atom 1.0 ----------

_atom_feed_mappings = (
    (("title",), "title"),
    (("link", "url"), "id"),
    (("description", "desc", "summary"), "subtitle"),
    (("pubDate", "pubdate", "date", "published", "updated"), "updated", lambda(x): _format_datetime("atom",x)),
    (("category",), "category"),
    (("author",), "author", lambda(x): _atomise_author(x))
)

_atom_item_mappings = (
    (("title",), "title"),
    (("link", "url"), "id"),
    (("link", "url"), "link", lambda(x): _atomise_link(x)),
    (("description", "desc", "summary"), "summary"),
    (("pubDate", "pubdate", "date", "published", "updated"), "updated", lambda(x): _format_datetime("atom",x)),
    (("category",), "category"),
    (("author",), "author", lambda(x): _atomise_author(x))
)

def _get_tz_offset():

    """
    Return the current timezone's offset from GMT as a string
    in the format +/-HH:MM, as required by RFC3339.
    """

    seconds = -1*timezone    # Python gets the offset backward! >:(
    minutes = seconds/60
    hours = minutes/60
    minutes = minutes - hours*60
    if seconds < 0:
        return "-%02d:%d" % (hours, minutes)
    else:
        return "+%02d:%d" % (hours, minutes)

def _convert_datetime(time):

    """
    Convert time, which may be one of a whole lot of things, into a
    standard 9 part time tuple.
    """

    if (type(time) is tuple and len(time) ==9) or type(time) is struct_time:
        # Already done!
        return time
    elif type(time) is int or type(time) is float:
        # Assume this is a seconds-since-epoch time
        return localtime(time)
    elif type(time) is str:    
        if time.isalnum():
            # String is alphanumeric - a time stamp?
            try:
                return strptime(time, "%a, %d %b %Y %H:%M:%S %Z")
            except ValueError:
                raise Exception("Unrecongised time format!")        
        else:
            # Maybe this is a string of an epoch time?
            try:
                return localtime(float(time))
            except ValueError:
                # Guess not.
                raise Exception("Unrecongised time format!")                 
    else:
        # No idea what this is.  Give up!
        raise Exception("Unrecongised time format!")

def _format_datetime(feed_type, time):

    """
    Convert some representation of a date and time into a string which can be
    used in a validly formatted feed of type feed_type.  Raise an
    Exception if this cannot be done.
    """
    
    # First, convert time into a time structure
    time = _convert_datetime(time)

    # Then, convert that to the appropriate string
    if feed_type is "rss2":
        return strftime("%a, %d %b %Y %H:%M:%S %Z", time)
    elif feed_type is "atom":
        return strftime("%Y-%m-%dT%H:%M:%S", time) + _get_tz_offset();

def _atomise_link(link):

    if type(link) is dict:
        return dict
    else:
        return {"href" : link}
        
def _atomise_author(author):

    """
    Convert author from whatever it is to a dictionary representing an
    atom:Person construct.
    """

    if type(author) is dict:
        return author
    else:
        if author.startswith("http://") or author.startswith("www"):
            # This is clearly a URI
            return {"uri" : author}
        elif "@" in author and "." in author:
            # This is most probably an email address
            return {"email" : author}
        else:
            # Must be a name
            return {"name" : author}

def _rssify_author(author):

    """
    Convert author from whatever it is to a plain old email string for
    use in an RSS 2.0 feed.
    """
    
    if type(author) is dict:
        try:
            return author["email"]
        except KeyError:
            return None
    else:
        if "@" in author and "." in author:
            # Probably an email address
            return author
        else:
            return None

def _add_subelems(root_element, mappings, dictionary):

    """
    Add one subelement to root_element for each key in dictionary
    which is supported by a mapping in mappings
    """

    for mapping in mappings:
        for key in mapping[0]:
            if key in dictionary:
                if len(mapping) == 2:
                    value = dictionary[key]
                elif len(mapping) == 3:
                    value = mapping[2](dictionary[key])
                _add_subelem(root_element, mapping[1], value)
                break                
    
def _add_subelem(root_element, name, value):

    if value is None:
        return

    if type(value) is dict:
        ### HORRIBLE HACK!
        if name=="link":
            ET.SubElement(root_element, name, href=value["href"])
        else:
            subElem = ET.SubElement(root_element, name)
            for key in value:
                _add_subelem(subElem, key, value[key])
    else:
        ET.SubElement(root_element, name).text = value

def _stringify(tree, pretty):

    """
    Turn an ElementTree into a string, optionally with line breaks and indentation.
    """

    if pretty and feedformatterCanPrettyPrint:
        string = StringIO()
        doc = FromXml(ET.tostring(tree))
        PrettyPrint(doc,string,indent="    ")
        return string.getvalue()
    else:
        return ET.tostring(tree)

class Feed:

    ### INTERNAL METHODS ------------------------------
    
    def __init__(self, feed=None, items=None):

        if feed:
            self.feed = feed
        else:
            self.feed = {}
        if items:
            self.items = items
        else:
            self.items = []
        self.entries = self.items

    ### RSS 1.0 STUFF ------------------------------
        
    def validate_rss1(self):

        """Raise an InvalidFeedException if the feed cannot be validly
        formatted as RSS 1.0."""

        # <channel> must contain "title"
        if "title" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            "RSS 1.0 feed must contain a title subelement")

        # <channel> must contain "link"
        if "link" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            " RSS 1.0 feeds must contain a link subelement")

        # <channel> must contain "description"
        if "description" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            "RSS 1.0 feeds must contain a description subelement")

        # Each <item> must contain "title" and "link"
        for item in self.items:
            if "title" not in item:
                raise InvalidFeedException("Each item element in an RSS 1.0 "
                "feed must contain a title subelement")
            if "link" not in item:
                raise InvalidFeedException("Each item element in an RSS 1.0 "
                "feed must contain a link subelement")
        
    def format_rss1_string(self, validate=True, pretty=False):

        """Format the feed as RSS 1.0 and return the result as a string."""

        if validate:
            self.validate_rss1()
        RSS1root = ET.Element( 'rdf:RDF', 
            {"xmlns:rdf" : "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
             "xmlns" : "http://purl.org/rss/1.0/"} )
        RSS1channel = ET.SubElement(RSS1root, 'channel',
            {"rdf:about" : self.feed["link"]})
        _add_subelems(RSS1channel, _rss1_channel_mappings, self.feed)
        RSS1contents = ET.SubElement(RSS1channel, 'items')
        RSS1contents_seq = ET.SubElement (RSS1contents, 'rdf:Seq')
        for item in self.items:
            ET.SubElement(RSS1contents_seq, 'rdf:li', resource=item["link"])
        for item in self.items:
            RSS1item = ET.SubElement (RSS1root, 'item',
                {"rdf:about" : item["link"]})
            _add_subelems(RSS1item, _rss1_item_mappings, item)
        return _stringify(RSS1root, pretty=pretty)

    def format_rss1_file(self, filename, validate=True, pretty=False):

        """Format the feed as RSS 1.0 and save the result to a file."""

        string = self.format_rss1_string(validate, pretty)
        fp = open(filename, "w")
        fp.write(string)
        fp.close()

    ### RSS 2.0 STUFF ------------------------------

    def validate_rss2(self):

        """Raise an InvalidFeedException if the feed cannot be validly
        formatted as RSS 2.0."""

        # <channel> must contain "title"
        if "title" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            "RSS 2.0 feed must contain a title subelement")

        # <channel> must contain "link"
        if "link" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            " RSS 2.0 feeds must contain a link subelement")

        # <channel> must contain "description"
        if "description" not in self.feed:
            raise InvalidFeedException("The channel element of an "
            "RSS 2.0 feeds must contain a description subelement")

        # Each <item> must contain at least "title" OR "description"
        for item in self.items:
            if not ("title" in item or "description" in item):
                raise InvalidFeedException("Each item element in an RSS 2.0 "
                "feed must contain at least a title or description subelement")

    def format_rss2_string(self, validate=True, pretty=False):

        """Format the feed as RSS 2.0 and return the result as a string."""

        if validate:
            self.validate_rss2()
        RSS2root = ET.Element( 'rss', {'version':'2.0'} )
        RSS2channel = ET.SubElement( RSS2root, 'channel' )
        _add_subelems(RSS2channel, _rss2_channel_mappings, self.feed)
        for item in self.items:            
            RSS2item = ET.SubElement ( RSS2channel, 'item' )
            _add_subelems(RSS2item, _rss2_item_mappings, item)
        return _stringify(RSS2root, pretty=pretty)

    def format_rss2_file(self, filename, validate=True, pretty=False):

        """Format the feed as RSS 2.0 and save the result to a file."""

        string = self.format_rss2_string(validate, pretty)
        fp = open(filename, "w")
        fp.write(string)
        fp.close()

    ### ATOM STUFF ------------------------------

    def validate_atom(self):

        """Raise an InvalidFeedException if the feed cannot be validly
        formatted as Atom 1.0."""

        # Must have at least one "author" element in "feed" OR at least
        # "author" element in each "entry".
        if "author" not in self.feed:
            for entry in self.entries:
                if "author" not in entry:
                    raise InvalidFeedException("Atom feeds must have either at "
                    "least one author element in the feed element or at least "
                    " one author element in each entry element")

    def format_atom_string(self, validate=True, pretty=False):

        """Format the feed as Atom 1.0 and return the result as a string."""

        if validate:
            self.validate_atom()
        AtomRoot = ET.Element( 'feed', {"xmlns":"http://www.w3.org/2005/Atom"} )
        _add_subelems(AtomRoot, _atom_feed_mappings, self.feed)
        for entry in self.entries:
            AtomItem = ET.SubElement ( AtomRoot, 'entry' )
            _add_subelems(AtomItem, _atom_item_mappings, entry)
        return _stringify(AtomRoot, pretty=pretty)

    def format_atom_file(self, filename, validate=True, pretty=False):

        """Format the feed as Atom 1.0 and save the result to a file."""

        string = self.format_atom_string(validate, pretty)
        fp = open(filename, "w")
        fp.write(string)
        fp.close()

class InvalidFeedException(Exception):

    pass

### FACTORY FUNCTIONS ------------------------------

def fromUFP(ufp):

    return Feed(ufp["feed"], ufp["items"])



# Module:   main
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Main

...
"""

import os
import sys

from mercurial.hgweb import hgweb

from circuits.app import Daemon
from circuits import Manager, Debugger

try:
    from circuits.web.apps import MemoryMonitor
except ImportError:
    MemoryMonitor = None

from circuits.web.wsgi import Application, Gateway
from circuits.web import Logger, Server, Sessions, Static

from root import Root
from config import Config
from env import Environment
from tools import CacheControl, Compression, ErrorHandler, SignalHandler

def main():
    config = Config()

    manager = Manager()

    if config.get("debug"):
        manager += Debugger(
            events=config.get("verbose"),
            file=config.get("errorlog"),
        )

    environ = Environment(config)

    manager += environ

    if config.get("sock") is not None:
        bind = config.get("sock")
    elif ":" in config.get("bind"):
        address, port = config.get("bind").split(":")
        bind = (address, int(port),)
    else:
        bind = (config.get("bind"), config.get("port"),)

    manager += (Server(bind)
            + Sessions()
            + Root(environ)
            + CacheControl(environ)
            + ErrorHandler(environ)
            + SignalHandler(environ)
    )

    if MemoryMonitor is not None:
        MemoryMonitor(channel="/memory").register(manager)

    if not config.get("disable-logging"):
        manager += Logger(file=config.get("accesslog", sys.stdout))

    if not config.get("disable-static"):
        manager += Static(docroot=os.path.join(config.get("theme"), "htdocs"))

    if not config.get("disable-hgweb"):
        manager += Gateway(hgweb(environ.storage.repo_path), "/+hg")

    if not config.get("disable-compression"):
        manager += Compression(environ)

    if config.get("daemon"):
        manager += Daemon(config.get("pidfile"))

    manager.run()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Module:   search
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Search and Indexing SUpport

...
"""

import re

from sqlalchemy import func, exists
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import Column, ForeignKey, Integer, Sequence, String

import schema
from i18n import _
from dbm import Base
from errors import NotFoundErr
from utils import external_link, extract_links

class Title(Base):

    __tablename__ = "titles"

    id = Column(Integer, Sequence("titles_id_seq"), primary_key=True)
    title = Column(String(50))

    def __init__(self, title):
        self.title = title

    def __repr__(self):
        return "<Title('%s')>" % self.title

class Word(Base):

    __tablename__ = "words"

    id = Column(Integer, Sequence("words_id_seq"), primary_key=True)
    word = Column(String(50), index=True)
    page = Column(Integer, ForeignKey("titles.title"), index=True)
    count = Column(Integer)

    def __init__(self, word, page, count):
        self.word = word
        self.page = page
        self.count = count

    def __repr__(self):
        return "<Word('%s', '%s', %d)>" % (self.word, self.page, self.count)

class Link(Base):

    __tablename__ = "links"

    id = Column(Integer, Sequence("links_id_seq"), primary_key=True)
    src = Column(Integer, ForeignKey("titles.title"), index=True)
    target = Column(Integer, ForeignKey("titles.title"), index=True)
    label = Column(String(50))
    number = Column(Integer)

    def __init__(self, src, target, label, number):
        self.src = src
        self.target = target
        self.label = label
        self.number = number

    def __repr__(self):
        return "<Link('%s', '%s', '%s', %d)>" % (self.src, self.target,
                self.label, self.number)

class WikiSearch(object):
    """
    Responsible for indexing words and links, for fast searching and
    backlinks. Uses a cache directory to store the index files.
    """

    word_pattern = re.compile(ur"""\w[-~&\w]+\w""", re.UNICODE)

    def __init__(self, db, lang, storage):
        self.db = db
        self.lang = lang
        self.storage = storage

        self.stop_words_re = re.compile(u'^('+u'|'.join(re.escape(_(
u"""am ii iii per po re a about above
across after afterwards again against all almost alone along already also
although always am among ain amongst amoungst amount an and another any aren
anyhow anyone anything anyway anywhere are around as at back be became because
become becomes becoming been before beforehand behind being below beside
besides between beyond bill both but by can cannot cant con could couldnt
describe detail do done down due during each eg eight either eleven else etc
elsewhere empty enough even ever every everyone everything everywhere except
few fifteen fifty fill find fire first five for former formerly forty found
four from front full further get give go had has hasnt have he hence her here
hereafter hereby herein hereupon hers herself him himself his how however
hundred i ie if in inc indeed interest into is it its itself keep last latter
latterly least isn less made many may me meanwhile might mill mine more
moreover most mostly move much must my myself name namely neither never
nevertheless next nine no nobody none noone nor not nothing now nowhere of off
often on once one only onto or other others otherwise our ours ourselves out
over own per perhaps please pre put rather re same see seem seemed seeming
seems serious several she should show side since sincere six sixty so some
somehow someone something sometime sometimes somewhere still such take ten than
that the their theirs them themselves then thence there thereafter thereby
therefore therein thereupon these they thick thin third this those though three
through throughout thru thus to together too toward towards twelve twenty two
un under ve until up upon us very via was wasn we well were what whatever when
whence whenever where whereafter whereas whereby wherein whereupon wherever
whether which while whither who whoever whole whom whose why will with within
without would yet you your yours yourself yourselves""")).split())
+ur')$|.*\d.*', re.U|re.I|re.X)

    def split_text(self, text, stop=True):
        """Splits text into words, removing stop words"""

        for match in self.word_pattern.finditer(text):
            word = match.group(0)
            if not (stop and self.stop_words_re.match(word)):
                yield word.lower()

    def count_words(self, words):
        count = {}
        for word in words:
            count[word] = count.get(word, 0)+1
        return count

    def title_id(self, title):
        try:
            return self.db.query(Title.id).filter(
                    Title.title==title).one().id
        except NoResultFound:
            self.db.add(Title(title))
            return self.db.query(Title.id).filter(
                    Title.title==title).one().id

    def update_words(self, title, text):
        title_id = self.title_id(title)
        self.db.query(Word).filter(Word.page==title_id).delete()
        if not text:
            return
        words = self.count_words(self.split_text(text))
        title_words = self.count_words(self.split_text(title))
        for word, count in title_words.iteritems():
            words[word] = words.get(word, 0) + count
        for word, count in words.iteritems():
            self.db.add(Word(word, title_id, count))

    def update_links(self, title, links_and_labels):
        title_id = self.title_id(title)
        self.db.query(Link).filter(Link.src==title_id).delete()
        for number, (link, label) in enumerate(links_and_labels):
            self.db.add(Link(title_id, link, label, number))

    def orphaned_pages(self):
        """Gives all pages with no links to them."""

        stmt = ~exists().where(Link.target==Title.title)
        orphaned = self.db.query(Title.title).\
                filter(stmt).\
                order_by(Title.title)
        for (title,) in orphaned:
            yield unicode(title)

    def wanted_pages(self):
        """Gives all pages that are linked to, but don't exist, together with
        the number of links."""

        stmt = ~exists().where(Title.title==Link.target)
        wanted = self.db.query(func.count(), Link.target).\
                filter(stmt).\
                group_by(Link.target).\
                order_by(-func.count())
        for refs, title, in wanted:
            title = unicode(title)
            if not external_link(title) and not title.startswith('+'):
                yield refs, title

    def page_backlinks(self, title):
        """Gives a list of pages linking to specified page."""

        backlinks = self.db.query(func.distinct(Title.title)).\
                join((Link, Link.src==Title.id)).\
                filter(Link.target==title).\
                order_by(Title.title)
        for (backlink,) in backlinks:
            yield unicode(backlink)

    def page_links(self, title):
        """Gives a list of links on specified page."""

        title_id = self.title_id(title)
        links = self.db.query(Link.target).\
                filter(Link.src==title_id).\
                order_by(Link.number)
        for link in links:
            yield unicode(link)

    def page_links_and_labels (self, title):
        title_id = self.title_id(title)
        links = self.db.query(Link.target, Link.label).\
                filter(Link.src==title_id).\
                order_by(Link.number)
        for link, label in links:
            yield unicode(link), unicode(label)

    def find(self, words):
        """Iterator of all pages containing the words, and their scores."""

        ranks = []
        for word in words:
            # Calculate popularity of each word.
            rank = self.db.query(func.sum(Word.count)).filter(
                    Word.word.like("%%%s%%" % word)).first()[0]
            # If any rank is 0, there will be no results anyways
            if not rank:
                return
            ranks.append((rank, word))
        ranks.sort()
        # Start with the least popular word. Get all pages that contain it.
        first_rank, first = ranks[0]
        rest = ranks[1:]
        first_counts = self.db.query(Word.page, Title.title,
                func.sum(Word.count)).\
                        filter(Word.word.like("%%%s%%" % first)).\
                        filter(Title.id==Word.page).\
                        group_by(Word.page)
        # Check for the rest of words
        for title_id, title, first_count in first_counts:
            # Score for the first word
            score = float(first_count)/first_rank
            for rank, word in rest:
                count = self.db.query(func.sum(Word.count)).\
                        filter(Word.page==title_id).\
                        filter(Word.word.like("%%%%s%%" % word)).\
                        first()
                if not count:
                    # If page misses any of the words, its score is 0
                    score = 0
                    break
                score += float(count)/rank
            if score > 0:
                yield int(100*score), unicode(title)

    def reindex_page(self, page, title, text=None):
        """Updates the content of the database, needs locks around."""

        if text is None:
            _get_text = getattr(page, '_get_text', lambda: u'')
            try:
                text = _get_text()
            except NotFoundErr:
                text = None
                title_id = self.title_id(title)
                if not list(self.page_backlinks(title)):
                    self.db.query(Title).filter(Title.id==title_id).delete()

        if text is not None:
            links = extract_links(text)
        else:
            links = []

        self.update_links(title, links)
        self.update_words(title, text or u'')

    def update_page(self, page, title, data=None, text=None):
        """Updates the index with new page content, for a single page."""

        if text is None and data is not None:
            text = unicode(data, self.storage.charset, 'replace')

        self.db.begin(subtransactions=True)

        try:
            self.set_last_revision(self.storage.repo_revision())
            self.reindex_page(page, title, text)
            self.db.commit()
        except:
            self.db.rollback()
            raise

    def reindex(self, environ, pages):
        """Updates specified pages in bulk."""

        self.db.begin(subtransactions=True)

        try:
            for title in pages:
                page = environ.get_page(title)
                self.reindex_page(page, title)
            self.db.commit()
        except Exception, e:
            self.db.rollback()
            raise

    def set_last_revision(self, rev):
        """Store the last indexed repository revision."""

        # We use % here because the sqlite3's substitiution doesn't work
        # We store revision 0 as 1, 1 as 2, etc. because 0 means "no revision"
        
        sysinfo = self.db.query(schema.System).get("search_revision")
        if sysinfo is None:
            self.db.add(schema.System("search_revision", (int(rev + 1))))
        else:
            sysinfo.value = int(rev + 1)

    def get_last_revision(self):
        """Retrieve the last indexed repository revision."""

        sysinfo = self.db.query(schema.System).get("search_revision")
        if sysinfo is None:
            return -1
        else:
            return int(sysinfo.value) - 1

    def update(self, environ):
        """Reindex al pages that changed since last indexing."""

        last_rev = self.get_last_revision()
        if last_rev == -1:
            changed = self.storage.all_pages()
        else:
            changed = self.storage.changed_since(last_rev)
        self.reindex(environ, changed)
        rev = self.storage.repo_revision()
        self.set_last_revision(rev)



# Module:   utils
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Utility Macros

Various Utility Macros
"""

from inspect import getdoc

def macros(macro, environ, data, *args, **kwargs):
    """Display a list of available macros and their documentation.
    
    **Arguments:** //No Arguments//

    **Example(s):**
    {{{
    <<macros>>
    }}}

    You're looking at it! :)
    """

    macros = environ.macros.items()
    s = "\n".join(["== %s ==\n%s\n" % (k, getdoc(v)) for k, v in macros])

    return environ.parser.generate(s, context="inline",
            environ=(environ, data))
